;;; slider-widget.el --- Interactive slider widget -*- lexical-binding: t; -*-

;; Copyright (c) 2018 Ian Dunn

;; Author: Ian Dunn <dunni@gnu.org>
;; Maintainer: Ian Dunn <dunni@gnu.org>
;; Keywords: Development, widgets
;; Version: 0.1
;; Package-Requires: ((emacs "25.1"))

;; This file is NOT part of GNU Emacs.

;; This program is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free Software
;; Foundation; either version 3, or (at your option) any later version.

;; This program is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;; details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;; Widget for a slider:
;;
;; --------[]--------------
;;
;; Create it like any other widget, using `widget-create' or the like.  The type
;; is 'slider.
;;
;; A slider has several properties:
;;
;; :background-string   String to display in an inactive part of the slider.  In
;;                      the above example, the string is "--" (default).
;;
;; :cursor-string       String for the cursor in the slider.  In the above
;;                      example, it's "[]" (default).
;;
;; :progress-set        Function to call when the value is interactively set.
;;                      By default, this is empty.
;;
;; :width               Number of "background-strings" to use to create the
;;                      slider.  This isn't necessarily the width in characters
;;                      of the slider.  The default is 15.
;;
;; The value of the slider is a float value between 0 and 1.

;;; Code:

(eval-when-compile (require 'subr-x))
(require 'wid-edit)

;; A few examples; b is the number of bg-strs before the cursor, a is the number
;; of strings after it.  Width is 10.

;;  0%: []------------------ (b: 0, a:9)
;;  5%: []------------------ (b: 0, a:9) (Rounding Down)
;; 10%: --[]---------------- (b: 1, a:8)
;; 20%: ----[]-------------- (b: 2, a:7)
;; 30%: ------[]------------ (b: 3, a:6)
;; 40%: --------[]---------- (b: 4, a:5)
;; 50%: ----------[]-------- (b: 5, a:4)

;; Compatibility for Emacs < 26.1
(unless (fboundp 'when-let*)
  (defalias 'when-let* 'when-let))

(defun widget-slider-value-to-internal (widget progress)
  "Convert float PROGRESS to a string for WIDGET."
  (let* ((bg-str (widget-get widget :background-string))
         (cursor-str (widget-get widget :cursor-string))
         (width (widget-get widget :width))
         (bg-before (floor (* progress width)))
         (bg-after (- width bg-before 1))) ;; -1 for the cursor
    (mapconcat
     'identity
     (append
      (make-list bg-before bg-str)
      (list cursor-str)
      (make-list bg-after bg-str))
     "")))

(defun widget-slider--convert (offset width bg-str-length)
  (let* ((bg-before (/ (float offset)
                       (float bg-str-length))))
    (/ bg-before (float width))))

(defun widget-slider-value-to-external (widget progress-string)
  (save-match-data
    (when-let* ((bg-str (widget-get widget :background-string))
                (cursor-str (widget-get widget :cursor-string))
                (width (widget-get widget :width))
                (string-to-match (format "\\(\\(?:%s\\)*\\)%s\\(\\(?:%s\\)*\\)"
                                         (regexp-quote bg-str)
                                         (regexp-quote cursor-str)
                                         (regexp-quote bg-str)))
                (pos (string-match string-to-match progress-string)))
      (widget-slider--convert (length (match-string 1 progress-string))
                              width
                              (length bg-str)))))

(defun widget-slider-notify (widget _changed &optional event)
  "Notification function for a slider widget, WIDGET.

EVENT is the event that triggered it, if any."
  (let* ((from (widget-get widget :from))
         (pos (if event
                  (widget-event-point event)
                (point)))
         (width (widget-get widget :width))
         (offset (- pos from))
         (progress-set (widget-get widget :progress-set))
         (bg-str (widget-get widget :background-string))
         (progress (widget-slider--convert offset width (length bg-str))))
    (widget-value-set widget progress)
    (when progress-set
      (funcall progress-set progress))))

;; mouse-down-action is set for compatibility with the termux interface, which
;; doesn't send mouse-click events, just mouse-down.
(define-widget 'slider 'item
  "Slider widget"
  :value-to-internal #'widget-slider-value-to-internal
  :value-to-external #'widget-slider-value-to-external
  :width 15
  :notify #'widget-slider-notify
  :mouse-down-action (lambda (_wid &optional _event) t)
  :background-string "--"
  :cursor-string "[]"
  :progress-set #'ignore
  :format "%[%v%]")

(provide 'slider-widget)

;;; slider-widget.el ends here
